import os
import sys
import json
import logging
import subprocess
from azure.cli.core import get_default_cli

logger = logging.getLogger("azureml_ngc.azutils")

def evaluate_az_cmd(az_cmd):
    print(az_cmd)
    cmd_proc = subprocess.run(az_cmd, shell = True, stdout=subprocess.PIPE, stderr = subprocess.PIPE)
    # reading output and error
    cmd_stdout = cmd_proc.stdout.decode("utf-8")
    cmd_details = json.loads(cmd_stdout)
    return cmd_details

def az_cli(args_str):
    
    args = args_str.split()
    cli = get_default_cli()
    cli.invoke(args)
    if cli.result.result:
        return cli.result.result
    elif cli.result.error:
        raise cli.result.error
    return True

def get_resource_type_values(resource_type,search_args,value='name'):
    search_cmd = 'az {} list{}'.format(resource_type,search_args)
    cmd_details = evaluate_az_cmd(search_cmd)
    return [cmd_detail[value] for cmd_detail in cmd_details]

def resource_exist(resource_type,resource_name,search_args):
    print("Searching for {}; Named: {}".format(resource_type,resource_name))
    return resource_name in get_resource_type_values(resource_type,search_args)

def get_resources_args(resource_group):
    priority = 500

    if resource_exist('network nsg','{0}-nsg'.format(resource_group),''):
        priorities = get_resource_type_values('network nsg rule',' --nsg-name {0}-nsg -g {0}'.format(resource_group),'priority')
        if priority in priorities:
           priority = max(priorities)+1 

    resources_args = {'vnet':{},'nsg':{},'nsg rule':{}}
    resources_args['vnet']['s']=''
    resources_args['nsg']['s']=''
    resources_args['nsg rule']['s']=' --nsg-name {0}-nsg -g {0}'.format(resource_group)
    resources_args['vnet']['c']=' --address-prefix 10.0.0.0/16 --subnet-name default --subnet-prefix 10.0.0.0/24'
    resources_args['nsg']['c']=''
    resources_args['nsg rule']['c'] = (
        ' --nsg-name {}-nsg --priority {} --source-address-prefixes Internet'
        ' --destination-port-ranges 8786 8787 8888 --destination-address-prefixes "*"'
        ' --access Allow --protocol Tcp  --description "Allow Internet to Dask on ports 8786 8787 8888."'
    ).format(resource_group, priority)
    return resources_args

def set_resource(resource_group,resource_type,resource_name,search_args,create_args):
    if resource_exist(resource_type,resource_name,search_args):
        print('{} was found\n'.format(resource_name))
    else:
        print('Creating {}\n'.format(resource_name))
        evaluate_az_cmd('az {} create -g {} -n {}{}'.format(resource_type,resource_group,resource_name,create_args))

def set_resources(resource_group,location):
    resources_args = get_resources_args(resource_group)
    for resource_type, args in resources_args.items():
        set_resource(resource_group,'network {}'.format(resource_type),
        '{}-{}'.format(resource_group,"_".join(resource_type.split())),args['s'],args['c'])
    return ['{}-{}'.format(resource_group,"_".join(x.split())) for x in resources_args.keys()]

def login():
    az_cli("login")
    return



from azureml.core import Workspace, Environment
from azureml.core.conda_dependencies import CondaDependencies
from azureml.core.compute import ComputeTarget, AmlCompute
from azureml.core import Experiment
from azureml.train.estimator import Estimator
import click, os, time, re
from azureml.exceptions._azureml_exception import ProjectSystemException

def get_ssh_keys():
    from cryptography.hazmat.primitives import serialization as crypto_serialization
    from cryptography.hazmat.primitives.asymmetric import rsa
    from cryptography.hazmat.backends import default_backend as crypto_default_backend
    dir_path = os.path.join(os.getcwd(), ".ssh")
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)
    pub_key_file = os.path.join(dir_path, "key.pub")
    pri_key_file = os.path.join(dir_path, "key")
    keys_exist = True
    if not os.path.exists(pub_key_file):
        print("Public SSH key does not exist!")
        keys_exist = False
    if not os.path.exists(pri_key_file):
        print("Private SSH key does not exist!")
        keys_exist = False
    if not keys_exist:
        key = rsa.generate_private_key(
            backend=crypto_default_backend(), public_exponent=65537, key_size=2048
        )
        private_key = key.private_bytes(
            crypto_serialization.Encoding.PEM,
            crypto_serialization.PrivateFormat.PKCS8,
            crypto_serialization.NoEncryption(),
        )
        public_key = key.public_key().public_bytes(
            crypto_serialization.Encoding.OpenSSH,
            crypto_serialization.PublicFormat.OpenSSH,
        )
        with open(pub_key_file, "wb") as f:
            f.write(public_key)
        with open(pri_key_file, "wb") as f:
            f.write(private_key)
        os.chmod(pri_key_file, 0o600)
    with open(pub_key_file, "r") as f:
        pubkey = f.read()
    return pubkey, pri_key_file

def createOrGetComputeTarget(
    ws, ct_name, vm_name, vm_priority, ssh_key_pub,min_nodes,max_nodes,idle_seconds,username
):
    config = AmlCompute.provisioning_configuration(
        vm_size=vm_name,
        min_nodes=min_nodes,
        max_nodes=max_nodes,
        vm_priority=vm_priority,
        idle_seconds_before_scaledown=idle_seconds,
        admin_username=username,
        admin_user_ssh_key=ssh_key_pub,
        remote_login_port_public_access="Enabled",
    )
    ct = ComputeTarget.create(ws, ct_name, config)
    ct.wait_for_completion(show_output=True)
    return ct

def createOrGetEnvironment(ws, env_name):
    environment_name = env_name
    python_interpreter = "/opt/miniconda/bin/python"
    conda_packages = ["matplotlib","jupyterlab"]
    ### CREATE OR RETRIEVE THE ENVIRONMENT
    if environment_name not in ws.environments:
        env = Environment(name=environment_name)
        env.docker.enabled = True
        env.docker.base_image = None
        env.docker.base_dockerfile = 'FROM mcr.microsoft.com/azureml/openmpi4.1.0-cuda11.1-cudnn8-ubuntu18.04\n RUN conda install -c r -y pip=20.1.1 && \\\n pip install jupyterlab && \\\n pip install nvidia-pyindex && \\\n pip install nvidia-tao'
        env.python.interpreter_path = python_interpreter
        env.python.user_managed_dependencies = True
        env.environment_variables = {"AZUREML_COMPUTE_USE_COMMON_RUNTIME":"false"}
        conda_dep = CondaDependencies()
        for conda_package in conda_packages:
            conda_dep.add_conda_package(conda_package)
        env.python.conda_dependencies = conda_dep
        env.register(workspace=ws)
    else:
        env = ws.environments[environment_name]
    return env

def setup_port_forwarding(run,ct,username):
    jupyter_address = run.get_metrics()["jupyter"]
    headnode_ip = run.get_metrics()["jupyter"].split(":")[0]
    headnode_public_ip = ct.list_nodes()[0]["publicIpAddress"]
    headnode_public_port = ct.list_nodes()[0]["port"]
    print("headnode_public_ip: {}".format(headnode_public_ip))
    print("headnode_public_port: {}".format(headnode_public_port))
    cmd = (
        "ssh -vvv -o StrictHostKeyChecking=no -N"
        f" -i {os.path.expanduser(self.admin_ssh_key)}"
        f" -L 0.0.0.0:{self.jupyter_port}:{headnode_ip}:8888"
    )
    cmd += f" {username}@{headnode_public_ip} -p {headnode_public_port}"
    portforward_proc = subprocess.Popen(
        cmd.split(),
        universal_newlines=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )

    #### Starting thread to keep the SSH tunnel open on Windows
    #self.portforward_logg = port_forward_logger(self.portforward_proc)
    #self.portforward_logg.start()

def get_jupyter_link(ws,run,jupyter_port):
    hostname = "localhost"
    location = ws.get_details()["location"]
    token = run.get_metrics()["token"]
    jupyter_url = f"http://{hostname}:{jupyter_port}/?token={token}"
    return jupyter_url

def setUpJupyterLab(ws,ct, env,exp_name,script_params,username,path,tags,jupyter_port):
    exp = Experiment(ws, exp_name)
    estimator = Estimator(path,
        compute_target=ct,
        entry_script="start_jupyter.py",
        environment_definition=env,
        script_params=script_params,
        node_count=1,  ### start only scheduler
        use_docker=True,
    )
    run = exp.submit(estimator, tags=tags)
    print("Waiting for compute cluster's IP")
    while (
        run.get_status() != "Canceled"
        and run.get_status() != "Failed"
        and "jupyter"
        not in run.get_metrics()  # and "scheduler" not in run.get_metrics()
    ):
        print(".", end="")
        time.sleep(5)
    if run.get_status() == "Canceled" or run.get_status() == "Failed":
        print("Failed to start the AzureML Compute Cluster")
        raise Exception("Failed to start the AzureML Compute Cluster.")
    print()
    print("Jupyter session is running...")
    print("\n\n")
    setup_port_forwarding(run,ct,username)
    jupyter_link = get_jupyter_link(ws,run,jupyter_port)
    print("Connections established")
    return run,jupyter_link




import os
import sys
import urllib.request
from tqdm import tqdm
import zipfile
import json
import logging
import threading
import subprocess
import tarfile
from progressbar import ProgressBar

logger = logging.getLogger("azureml_ngc.ngc_content")

def getRootFolder(zip_ref):
    members = zip_ref.infolist()
    rootFolders = set([member.filename.split("/")[0] for member in members])
    if len(rootFolders) == 1:
        rootFolder = rootFolders.pop()
    else:
        rootFolder = ""
    return rootFolder

def pruneTargetFolder(target_folder,zip_ref):
    rootFolder = getRootFolder(zip_ref)
    if rootFolder!="":
        subfolders = target_folder.split("/")
        if len(subfolders) > 1:
            if subfolders[-1] == rootFolder:
                del subfolders[-1]
                target_folder = "/".join(subfolders)
    return target_folder

def downloadurltofile(url, filename, orig_filename):
    if not os.path.exists(filename):
        logger.info(f"    --> [DOWNLOAD] Downloading {orig_filename} to {filename}<--")
        with open(filename, "wb") as file:
            with urllib.request.urlopen(url) as resp:
                length = int(resp.getheader("content-length"))
                blocksize = max(4096, length // 100)
                with tqdm(total=length, file=sys.stdout) as pbar:
                    while True:
                        buff = resp.read(blocksize)
                        if not buff:
                            break
                        file.write(buff)
                        pbar.update(len(buff))
        logger.info(f"    -->> [DOWNLOAD] File {orig_filename} downloaded... <<--")
    else:
        logger.info(
            f"    -->> [DOWNLOAD] {orig_filename} file already exists locally <<--"
        )


def download(url, targetfolder, targetfile):
    path = os.getcwd()
    data_dir = os.path.abspath(os.path.join(path, targetfolder))

    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    targetfile_fullpath = os.path.join(data_dir, targetfile)
    downloadurltofile(url, targetfile_fullpath, targetfile)
    return data_dir, targetfile


def makedirs(path, targetfolder,data_dir):
    subfolders = targetfolder.split('/')
    if len(subfolders) == 1:
        os.makedirs(data_dir)
    else:
        for subfolder in subfolders:
            path =os.path.abspath(os.path.join(path, subfolder))
            if not os.path.exists(path):
                os.makedirs(path)

def unzipFile(filename, srcfolder, destfolder):
    path = os.getcwd()
    data_dir = os.path.abspath(os.path.join(path, srcfolder))
    filepath = os.path.abspath(os.path.join(data_dir, filename))
    destfolder = os.path.join(data_dir, destfolder)

    if not os.path.exists(destfolder):
        logger.info(f"    -->> [EXTRACT] Extracting {filename} to {destfolder}... <<--")
        with zipfile.ZipFile(filepath, "r") as zip_ref:
            destfolder = pruneTargetFolder(destfolder,zip_ref)
            zip_ref.extractall(destfolder)
    else:
        logger.info(
            f"    -->> [EXTRACT] {filename} already extracted to {destfolder}... <<--"
        )

def show_progress(count, block_size, total_size):
    global pbar
    global processed
    
    if count == 0:
        pbar = ProgressBar(maxval=total_size)
        processed = 0
    
    processed += block_size
    processed = min(processed,total_size)
    pbar.update(processed)

def decompress_tarfile(filename, srcfolder, destfolder):
    path = os.getcwd()
    data_dir = os.path.abspath(os.path.join(path, srcfolder))
    filepath = os.path.abspath(os.path.join(data_dir, filename))
    destfolder = os.path.join(data_dir, destfolder)   

    if not os.path.exists(destfolder):

        logger.info(f"    -->> [EXTRACT] Extracting {filename} to {destfolder}... <<--")
        
        tar = tarfile.open(filepath)
        print("...Getting information from {0} about files to decompress".format(filename))
        members = tar.getmembers()
        numFiles = len(members)
        so_far = 0
        for member_info in members:
            tar.extract(member_info,path=destfolder)
            show_progress(so_far, 1, numFiles)
            so_far += 1
        pbar.finish()
        print("...All {0} files have been decompressed".format(numFiles))
        tar.close()
    else:
        logger.info(
            f"    -->> [EXTRACT] {filename} already extracted to {destfolder}... <<--"
        )



def upload_data(workspace, datastore, src_dir, tgt_path, overwrite=False):
    print(src_dir, tgt_path)
    datastore.upload(
        src_dir=src_dir, target_path=tgt_path, show_progress=True, overwrite=overwrite
    )
    logger.info(
        f"    -->> [UPLOAD] Completed uploading folder {src_dir} to {tgt_path} in {datastore.name}... <<--"
    )

def get_config(configfile):
    jsonfile = open(configfile)
    configdata = json.load(jsonfile)
    return configdata

def flush(proc, proc_log):
    while True:
        proc_out = proc.stdout.readline()
        if proc_out == "" and proc.poll() is not None:
            proc_log.close()
            break
        elif proc_out:
            sys.stdout.write(proc_out)
            proc_log.write(proc_out)
            proc_log.flush()

def evaluate_cmd(cmd, logFileName):
    cmd_log = open(logFileName, "a")
    cmd_proc = subprocess.Popen(
        cmd.split(),
        universal_newlines=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    cmd_flush = threading.Thread(target=flush, args=(cmd_proc, cmd_log))
    cmd_flush.start()
    flush(cmd_proc, cmd_log)
    return cmd_proc

def validate_path(datadir,destfolder):
    subfolders = destfolder.split('/')
    if(len(subfolders)>1):
        current_dir = datadir
        for subfolder in subfolders[0:-1]:
            current_dir = os.path.join(current_dir, subfolder)
            if not os.path.exists(current_dir):
                os.makedirs(current_dir)

def clone_github_repo(github_url,datadir, destfolder):
    path = os.getcwd()
    data_dir = os.path.abspath(os.path.join(path, datadir))

    if not os.path.exists(data_dir):
        os.makedirs(data_dir)
    
    validate_path(datadir,destfolder)
    logFileName = os.path.join(data_dir, '{}_upload_log.txt'.format(destfolder))
    destfolder = os.path.join(data_dir, destfolder)
    if os.path.exists(destfolder):
        print("Git Clone Destination folder already exists")
    else:
        print("Cloning")
        cmd = 'git clone {} {}'.format(github_url,destfolder)
        proc = evaluate_cmd(cmd, logFileName)
        proc.kill()




# Imports
import os
import sys
import uuid
import time
import socket
import argparse
import threading
import subprocess
import logging
import nbformat as nbf


def validate_path(datadir,destfolder):
    destfolder = destfolder.strip("/")
    subfolders = destfolder.split('/')
    if(len(subfolders)>1):
        current_dir = datadir
        for subfolder in subfolders:
            current_dir = os.path.join(current_dir, subfolder)
            if not os.path.exists(current_dir):
                os.makedirs(current_dir)

####---from mpi4py import MPI
def createCopyDataNotebook(fname,folders):
    nb = nbf.v4.new_notebook()
    text = """\
# Copy Required Data from Datastore."""
    code1 = """\
import azureml.core
from azureml.core import Workspace, Datastore

ws = Workspace.from_config()
datastore = ws.get_default_datastore()"""

    code_entries = ["datastore.download('./', '{0}', overwrite=False, show_progress=True)".format(x) for x in folders] 
    
    nb['cells'] = [nbf.v4.new_markdown_cell(text),
               nbf.v4.new_code_cell(code1)]
    
    nb['cells'].extend([nbf.v4.new_code_cell(code) for code in code_entries])
    
    with open(fname, 'w') as f:
        nbf.write(nb, f)

def createUploadDataNotebook(fname,folder):
    nb = nbf.v4.new_notebook()
    text = """\
# Upload Data to Datastore."""
    code1 = """\
import azureml.core
from azureml.core import Workspace, Datastore

ws = Workspace.from_config()
datastore = ws.get_default_datastore()"""

    nb['cells'] = [nbf.v4.new_markdown_cell(text),
               nbf.v4.new_code_cell(code1)]
    
    nb['cells'].append(nbf.v4.new_code_cell("datastore.upload('./', '{0}', overwrite=False, show_progress=True)".format(folder)))

    with open(fname, 'w') as f:
        nbf.write(nb, f)

def install(package,upgrade=False):
    if upgrade:
        subprocess.check_call([sys.executable, "-m", "pip", "install", package,"--upgrade"])
    else:
        subprocess.check_call([sys.executable, "-m", "pip", "install", package])

def flush(proc, proc_log):
    while True:
        proc_out = proc.stdout.readline()
        if proc_out == "" and proc.poll() is not None:
            proc_log.close()
            break
        elif proc_out:
            sys.stdout.write(proc_out)
            proc_log.write(proc_out)
            proc_log.flush()

def evaluate_cmd(cmd, logFileName):
    cmd_log = open(logFileName, "a")
    cmd_proc = subprocess.Popen(
        cmd.split(),
        universal_newlines=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )

    cmd_flush = threading.Thread(target=flush, args=(cmd_proc, cmd_log))
    cmd_flush.start()
    flush(cmd_proc, cmd_log)
    return cmd_proc

def mount(dataset,mnt_root,rel_mnt_path):
    mounted_path = os.path.join(mnt_root,rel_mnt_path)
    mount_context = dataset.mount(mounted_path)
    mount_context.start()
    return mounted_path

if __name__ == "__main__":
    logger = logging.getLogger(__name__)
    install("azureml-sdk")
    install("notebook")

    from azureml.core import Run
    from notebook.notebookapp import list_running_servers

    ### PARSE ARGUMENTS
    parser = argparse.ArgumentParser()
    parser.add_argument("--jupyter_token", default=uuid.uuid1().hex)
    parser.add_argument("--jupyter_port", default=8888)
    parser.add_argument("--use_gpu", default=False)
    parser.add_argument("--n_gpus_per_node", default=0)
    parser.add_argument('--output_path', dest='output_path', required=True)
    parser.add_argument('--ws_config', dest='ws_config', required=True)
    parser.add_argument('--folders', dest='folders', required=True)
    args, unparsed = parser.parse_known_args()

    ### CONFIGURE GPU RUN
    GPU_run = args.use_gpu

    if GPU_run:
        n_gpus_per_node = eval(args.n_gpus_per_node)

    ip = socket.gethostbyname(socket.gethostname())

    data = {
        "jupyter": ip + ":" + str(args.jupyter_port),
        "token": args.jupyter_token,
    }

    jupyter = data["jupyter"]
    token = data["token"]

    logger.debug("- args: ", args)
    logger.debug("- unparsed: ", unparsed)
    logger.debug("- my ip is ", ip)

    ####---# if rank == 0:
    running_jupyter_servers = list(list_running_servers())
    logger.debug("- running jupyter servers", running_jupyter_servers)

    ### if any jupyter processes running
    ### KILL'EM ALL!!!
    if len(running_jupyter_servers) > 0:
        for server in running_jupyter_servers:
            os.system(f'kill {server["pid"]}')

    ### RECORD LOGS
    run = Run.get_context()

    run.log("jupyter", jupyter)
    run.log("token", token)

    workspace_name = run.experiment.workspace.name.lower()
    run_id = run.get_details()["runId"]

    #mount_point = f"/mnt/batch/tasks/shared/LS_root/jobs/{workspace_name}/azureml/{run_id.lower()}/mounts/"
    mount_point = os.path.dirname(args.output_path)
    os.makedirs(mount_point, exist_ok=True)
    print("mount_point : " + mount_point)

    cmd = (
        f" jupyter lab --ip 0.0.0.0 --port {args.jupyter_port}"
        f" --NotebookApp.token={token}"
    )
    cmd += f" --notebook-dir={mount_point}"
    cmd += f" --allow-root --no-browser"

    jupyter_log = open("jupyter_log.txt", "a")
    jupyter_proc = subprocess.Popen(
        cmd.split(),
        universal_newlines=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    
    f = open(os.path.join(mount_point, "config.json"), "w")
    f.write(args.ws_config)
    f.close()

    fname = os.path.join(mount_point, 'CopyData.ipynb')
    folders = args.folders.split()
    createCopyDataNotebook(fname,folders)

    for folder in folders:
        validate_path(mount_point,folder)
        fname = os.path.join(mount_point, folder)
        f = open(os.path.join(fname,"config.json"), "w")
        f.write(args.ws_config)
        f.close()
        fname = os.path.join(fname, 'UploadData.ipynb')
        createUploadDataNotebook(fname,folder)
    
    #datastore = run.experiment.workspace.get_default_datastore()
    #rel_datastore_path = "clara"
    #datastore_paths = [(datastore, rel_datastore_path)]
    #dataset = Dataset.File.from_files(path=datastore_paths)
    #print("Mounting " + rel_datastore_path + " on " + mount_point)
    #mount(dataset,mount_point,rel_datastore_path)

    jupyter_flush = threading.Thread(target=flush, args=(jupyter_proc, jupyter_log))
    jupyter_flush.start()

    while not list(list_running_servers()):
        time.sleep(5)

    jupyter_servers = list(list_running_servers())
    assert len(jupyter_servers) == 1, "more than one jupyter server is running"

    flush(jupyter_proc, jupyter_log)

    if jupyter_proc:
        jupyter_proc.kill()

    run.complete()
    run.cancel()



import threading


class port_forward_logger(threading.Thread):
    def __init__(self, portforward_proc):
        super().__init__()

        self.portforward_proc = portforward_proc
        self.portforward_log = open("portforward_out_log.txt", "w")
        self.stop_logging = threading.Event()
        self.running = True

    def run(self):
        while self.running:
            portforward_out = self.portforward_proc.stdout.readline()
            if self.portforward_proc != "":
                self.portforward_log.write(portforward_out)
                self.portforward_log.flush()

        return

    def join(self, timeout=None):
        self.running = False
        self.stop_logging.set()
        super().join(timeout)



import logging

from azureml.core import Workspace, Environment
from azureml.core.conda_dependencies import CondaDependencies
from azureml.core.compute import ComputeTarget, AmlCompute
import click, os, time, re
from azureml_ngc_tools.AzureMLComputeCluster import AzureMLComputeCluster
from azureml_ngc_tools.cli import ngccontent
#from azureml_ngc_tools.cli import azcli_utils
from azureml.exceptions._azureml_exception import ProjectSystemException
#from dask_cloudprovider.azure import AzureVMCluster
#from dask.distributed import Client

### SETUP LOGGING
fileFormatter = logging.Formatter("%(asctime)s [%(levelname)-5.10s]  %(message)s")
logger = logging.getLogger("azureml_ngc")
logger.setLevel("DEBUG")

fileHandler = logging.FileHandler("azureml_ngc_tools.log", mode="w")
fileHandler.setFormatter(fileFormatter)
logger.addHandler(fileHandler)

def log(message,local=True):
    logger.info(message)
    if (local):
        print(message)

@click.command()
@click.option(
    "--login", is_flag=False, required=True, help="Path to the login config file"
)
@click.option("--app", is_flag=False, required=True, help="Path to the config file")
@click.version_option()
def start(login, app):
    install_packages()
    login_config = ngccontent.get_config(login)
    app_config = ngccontent.get_config(app)
    ### WORKSPACE
    ws = getWorkspace(login_config,logger)

    ### GPU RUN INFO
    gpus_per_node = validateVMSize(ws,login_config,logger)

    useDask = figureDaskUsage(login_config)
    folders = []
    ### UPLOAD ADDITIONAL CONTENT IF NOT EXISTS
    for additional_content in app_config["additional_content"]["list"]:
        url = additional_content["url"]
        targetfile = additional_content["filename"]
        src_path = additional_content["localdirectory"]
        dest_path = additional_content["computedirectory"]
        if (useDask):
            load_content_to_scheduler(url,targetfile,src_path,dest_path,additional_content,app_config,client)
        else:
            folder = load_content(url,targetfile,src_path,dest_path,additional_content,app_config,ws)
            folders.append(folder)
    
    log("folders {}".format(" ".join(folders)))

    if (not useDask):
        ct, pri_key_file = setUpComputeTarget(ws,login_config)
        env = createOrGetEnvironment(ws, login_config, app_config)
        cluster, jupyter_link = getAzureMLComputeCluster(ws,ct,env,login_config,pri_key_file,folders)
    else:
        azcli_utils.login()
        log("\nCreating a Dask Cluster with {} nodes".format(login_config["aml_compute"]["max_nodes"]))
        cluster, jupyter_link, client = getAzureVMCluster(ws,gpus_per_node,login_config,app_config)
    
    if (useDask):
        client.close()
    
    log(f"\n    Go to: {jupyter_link}")
    log("    Press Ctrl+C to stop the cluster.")

    try:
        while True:
            pass
    except KeyboardInterrupt:
        log("\nClosing Cluster ...")
        cluster.close()

def install_packages():
    import subprocess
    import sys
    packages = ["azure.cli.core"]
    for package in packages:
        subprocess.check_call([sys.executable, "-m", "pip", "install", package])

def figureDaskUsage(login_config):
    useDask = False
    if (
        login_config["aml_compute"]["max_nodes"]>1 
        or (login_config["aml_compute"]["max_nodes"]==1
        and 'dask' in login_config["aml_compute"]
        and login_config["aml_compute"]['dask']==True)
        ):
        useDask = True
    return useDask


def validateVMSize(ws,login_config,logger):
    vm_name = login_config["aml_compute"]["vm_name"].lower()
    workspace_vm_sizes = AmlCompute.supported_vmsizes(ws)
    pascal_volta_pattern = pattern = re.compile(
        r"[a-z]+_nc[0-9]+[s]?_v[2,3]"
    )  ### matches NC-series v2 and v3
    workspace_vm_sizes = [
        (e["name"].lower(), e["gpus"])
        for e in workspace_vm_sizes
        if pattern.match(e["name"].lower())
    ]
    workspace_vm_sizes = dict(workspace_vm_sizes)
    gpus_per_node = -1
    ### GET NUMBER OF GPUS PER VM
    if vm_name in workspace_vm_sizes:
        gpus_per_node = workspace_vm_sizes[vm_name]
        log("\n    VM SIze {} was recognized\n".format(vm_name))
    else:
        logger.exception("Unsupported vm_size {vm_size}".format(vm_size=vm_name))
        logger.exception("The specified vm size must be one of ...")

        for azure_gpu_vm_size in workspace_vm_sizes.keys():
            logger.exception("... " + azure_gpu_vm_size)
        raise Exception(
            "{vm_size} does not have Pascal or above GPU Family".format(vm_size=vm_name)
        )

    verify = f"""
    Number of GPUs per node: {gpus_per_node}
    """
    log(verify)

    return gpus_per_node

def getWorkspace(login_config,logger):
    subscription_id = login_config["azureml_user"]["subscription_id"]
    resource_group = login_config["azureml_user"]["resource_group"]
    workspace_name = login_config["azureml_user"]["workspace_name"]

    try:
        ws = Workspace(
            workspace_name=workspace_name,
            subscription_id=subscription_id,
            resource_group=resource_group,
        )
    except ProjectSystemException:
        msg = f'\n\nThe workspace "{workspace_name}" does not exist. '
        msg += f"Go to \n\n  "
        msg += f"-->> https://docs.microsoft.com/en-us/azure/machine-learning/how-to-manage-workspace <<--\n\n"
        msg += f"and create the workspace first.\n\n\n"
        msg += f"Your current configuration: \n\n"
        msg += f"Workspace name: {workspace_name} \n"
        msg += f"Subscription id: {subscription_id} \n"
        msg += f"Resource group: {resource_group}\n\n"
        logger.exception(msg)
        raise Exception(msg)

    ws.write_config(path="./", file_name="config.json")
    verify = f"""
    Subscription ID: {subscription_id}
    Resource Group: {resource_group}
    Workspace: {workspace_name}"""
    log(verify)

    return ws

def getAzureVMCluster(ws,gpus_per_node,login_config,app_config):
    location = ws.location
    resource_group = login_config["azureml_user"]["resource_group"]
    resources = azcli_utils.set_resources(resource_group,location)
    vnet = resources[0]
    security_group = resources[1]
    EXTRA_PIP_PACKAGES = login_config["aml_compute"]["pip_packages"]
    n_workers=int(login_config["aml_compute"]["max_nodes"])

    verify = f"""
    docker_image: {app_config["base_dockerfile"]}
    vm_size: {login_config["aml_compute"]["vm_name"]}
    resource_group: {resource_group}
    vnet: {vnet}
    security_group: {security_group}
    location: {location}"""
    log(verify)

    log(AzureVMCluster.get_cloud_init(
        resource_group=resource_group,
        location = location,
        vnet=vnet,
        security_group=security_group,
        n_workers=n_workers,
        vm_size=login_config["aml_compute"]["vm_name"],
        docker_image=app_config["base_dockerfile"],
        docker_args="--privileged",
        debug=True,
        disk_size=100, 
        auto_shutdown=False,
        security=False,
        env_vars={"EXTRA_PIP_PACKAGES": EXTRA_PIP_PACKAGES},
        worker_class="dask_cuda.CUDAWorker"))
    
    cluster = AzureVMCluster(
        resource_group=resource_group,
        location = location,
        vnet=vnet,
        security_group=security_group,
        n_workers=n_workers,
        vm_size=login_config["aml_compute"]["vm_name"],
        docker_image=app_config["base_dockerfile"],
        docker_args="--privileged",
        debug=True,
        disk_size=100, 
        auto_shutdown=False,
        security=False,
        env_vars={"EXTRA_PIP_PACKAGES": EXTRA_PIP_PACKAGES},
        worker_class="dask_cuda.CUDAWorker")
        
    log("All {} workers have been created\n".format(n_workers))

    client = Client(cluster)
    num_expected_total_gpus = n_workers*gpus_per_node

    log("Setting {} workers on the cluster for a total of {} GPUs\n".format(n_workers,num_expected_total_gpus))
    cluster.scale(n_workers)
    client = Client(cluster)
    
    num_provisioned_gpus = -1
    
    while (not len(client.nthreads().keys()) == num_expected_total_gpus):
        if(len(client.nthreads().keys())==num_provisioned_gpus):
            print(".", end="")
            time.sleep(5)
        else:
            num_provisioned_gpus = len(client.nthreads().keys())
            print("\nNumber of GPUs added to the cluster {} out of {}".format(num_provisioned_gpus,num_expected_total_gpus))
    print("\nNumber of GPUs added to the cluster {} out of {}".format(len(client.nthreads().keys()),num_expected_total_gpus))
    print('All requested {} GPUs have been added to the cluster'.format(num_expected_total_gpus))
    
    jupyter_link = "{}:8888".format(cluster.scheduler_address.split(':')[1].split('//')[1])
    return cluster, jupyter_link, client

def getAzureMLComputeCluster(ws,ct,env,login_config,pri_key_file,folders):
    cluster = AzureMLComputeCluster(
        workspace=ws,
        compute_target=ct,
        initial_node_count=1,
        experiment_name=login_config["aml_compute"]["exp_name"],
        environment_definition=env,
        jupyter_port=login_config["aml_compute"]["jupyter_port"],
        telemetry_opt_out=login_config["azureml_user"]["telemetry_opt_out"],
        admin_username=login_config["aml_compute"]["admin_name"],
        admin_ssh_key=pri_key_file,
        folders=folders,
        )
    jupyter_link = cluster.jupyter_link
    return cluster, jupyter_link


def setUpComputeTarget(ws,login_config):
    ### experiment name
    exp_name = login_config["aml_compute"]["exp_name"]
    ### azure ml names
    ct_name = login_config["aml_compute"]["ct_name"]
    vm_priority = login_config["aml_compute"]["vm_priority"]
    vm_name = login_config["aml_compute"]["vm_name"].lower()

    ### trust but verify
    verify = f"""
    Experiment name: {exp_name}"""
    log(verify)

    verify = f"""
    Compute target: {ct_name}
    VM Size: {vm_name}
    Priority: {vm_priority}
    """
    log(verify)

    ### get SSH keys
    ssh_key_pub, pri_key_file = get_ssh_keys()

    if ct_name not in ws.compute_targets:
        logger.warning(f"Compute target {ct_name} does not exist...")
        ct = createOrGetComputeTarget(
            ws, ct_name, vm_name, vm_priority, ssh_key_pub, login_config
        )
    else:
        ct = ws.compute_targets[ct_name]

        if ct.provisioning_state == "Failed":
            logger.warning(
                f"Compute target {ct_name} found but provisioning_state is showing as 'failed'..."
            )
            logger.warning(f"Deleting {ct_name} target and will attempt again...")
            logger.warning(
                 f"If this fails again check that you have enough resources in your subscription..."
            )

            ct.delete()
            time.sleep(5)
            ct = createOrGetComputeTarget(
                ws, ct_name, vm_name, vm_priority, ssh_key_pub, login_config
            )
        else:
            log(f"    Using pre-existing compute target {ct_name}")
    return ct, pri_key_file

def get_ssh_keys():
    from cryptography.hazmat.primitives import serialization as crypto_serialization
    from cryptography.hazmat.primitives.asymmetric import rsa
    from cryptography.hazmat.backends import default_backend as crypto_default_backend

    dir_path = os.path.join(os.getcwd(), ".ssh")

    if not os.path.exists(dir_path):
        os.makedirs(dir_path)

    pub_key_file = os.path.join(dir_path, "key.pub")
    pri_key_file = os.path.join(dir_path, "key")

    keys_exist = True

    if not os.path.exists(pub_key_file):
        print("Public SSH key does not exist!")
        keys_exist = False

    if not os.path.exists(pri_key_file):
        print("Private SSH key does not exist!")
        keys_exist = False

    if not keys_exist:
        key = rsa.generate_private_key(
            backend=crypto_default_backend(), public_exponent=65537, key_size=2048
        )

        private_key = key.private_bytes(
            crypto_serialization.Encoding.PEM,
            crypto_serialization.PrivateFormat.PKCS8,
            crypto_serialization.NoEncryption(),
        )
        public_key = key.public_key().public_bytes(
            crypto_serialization.Encoding.OpenSSH,
            crypto_serialization.PublicFormat.OpenSSH,
        )

        with open(pub_key_file, "wb") as f:
            f.write(public_key)

        with open(pri_key_file, "wb") as f:
            f.write(private_key)

        os.chmod(pri_key_file, 0o600)

    with open(pub_key_file, "r") as f:
        pubkey = f.read()

    return pubkey, pri_key_file


def createOrGetComputeTarget(
    ws, ct_name, vm_name, vm_priority, ssh_key_pub, login_config
):
    config = AmlCompute.provisioning_configuration(
        vm_size=vm_name,
        min_nodes=login_config["aml_compute"]["min_nodes"],
        max_nodes=login_config["aml_compute"]["max_nodes"],
        vm_priority=vm_priority,
        idle_seconds_before_scaledown=login_config["aml_compute"][
            "idle_seconds_before_scaledown"
        ],
        admin_username=login_config["aml_compute"]["admin_name"],
        admin_user_ssh_key=ssh_key_pub,
        remote_login_port_public_access="Enabled",
    )
    ct = ComputeTarget.create(ws, ct_name, config)
    ct.wait_for_completion(show_output=True)

    if ct.provisioning_state != "Succeeded":
        msg = f"Failed to create the cluster..."
        logger.exception(msg)
        raise Exception(msg)
    return ct


def createOrGetEnvironment(ws, login_config, app_config):
    environment_name = login_config["aml_compute"]["environment_name"]
    python_interpreter = login_config["aml_compute"]["python_interpreter"]
    conda_packages = login_config["aml_compute"]["conda_packages"]

    ### CREATE OR RETRIEVE THE ENVIRONMENT
    if environment_name not in ws.environments:
        log(f"Creating {environment_name} environment...")
        env = Environment(name=environment_name)
        env.docker.enabled = login_config["aml_compute"]["docker_enabled"]
        env.docker.base_image = None
        env.docker.base_dockerfile = f'FROM {app_config["base_dockerfile"]}'
        env.python.interpreter_path = python_interpreter
        env.python.user_managed_dependencies = True
        env.environment_variables = {"AZUREML_COMPUTE_USE_COMMON_RUNTIME":"false"}
        conda_dep = CondaDependencies()

        for conda_package in conda_packages:
            conda_dep.add_conda_package(conda_package)

        env.python.conda_dependencies = conda_dep
        env.register(workspace=ws)
    else:
        log(f"    Environment {environment_name} found...")
        env = ws.environments[environment_name]

    return env

def load_content(url,targetfile,src_path,dest_path,additional_content,app_config,ws):
    tarextensions = [".tar",".tbz",".tgz",".txz"]
    folder = ""
    if app_config["additional_content"]["download_content"]:
        if (
            "source" in additional_content.keys()
            and additional_content["source"] == "github"
        ):
            ngccontent.clone_github_repo(url,"additional_content",src_path)
            if "githubdirectory" in additional_content.keys():
                src_path = additional_content["githubdirectory"]
        else:
            lcl_path = "additional_content"
            if type(targetfile) == list:
                targetfiles = targetfile
                lcl_path = os.path.join(lcl_path,src_path)
                for targetfile in targetfiles:
                    if app_config["additional_content"]["download_content"]:
                        ngccontent.download(url+targetfile, lcl_path, targetfile)
            else:
                ngccontent.download(url, lcl_path, targetfile)
                if (
                    app_config["additional_content"]["unzip_content"]
                    and additional_content["zipped"]
                ):
                    fileroot, file_extension = os.path.splitext(targetfile)
                    if file_extension in tarextensions:
                        ngccontent.decompress_tarfile(targetfile, "additional_content", src_path)
                    else:
                        ngccontent.unzipFile(targetfile, "additional_content", src_path)
                    
    if app_config["additional_content"]["upload_content"]:
        print("Uploading content locally on: {}".format(src_path))
        print("To {} on datastore".format(dest_path))
        ngccontent.upload_data(
            ws,
            ws.get_default_datastore(),
            "additional_content/" + src_path,
            dest_path,
        )
        folder=dest_path

    return folder


def load_content_to_scheduler(url,targetfile,src_path,dest_path,additional_content,app_config,client):
    if app_config["additional_content"]["download_content"]:
        if (
            "source" in additional_content.keys()
            and additional_content["source"] == "github"
        ):
            #ngccontent.clone_github_repo(url,"additional_content",src_path)
            if "githubdirectory" in additional_content.keys():
                src_path = additional_content["githubdirectory"]
        else:
            lcl_path = src_path
            if type(targetfile) == list:
                targetfiles = targetfile
                lcl_path = dest_path
                for targetfile in targetfiles:
                    if app_config["additional_content"]["download_content"]:
                        client.run_on_scheduler(ngccontent.download,url+targetfile, lcl_path, targetfile)
            else:
                client.run_on_scheduler(ngccontent.download,url, lcl_path, targetfile) 
                if (
                    app_config["additional_content"]["unzip_content"]
                    and additional_content["zipped"]
                ):
                    log("Uploading {} content to folder {} on the scheduler".format(url,dest_path))
                    client.run_on_scheduler(ngccontent.unzipFile,targetfile, src_path, dest_path)      
    return

def go():
    start()


if __name__ == "__main__":
    go()
